@@ -5,15 +5,25 @@ use std::collections::HashMap;
 use std::sync::Arc;
 use std::time::Duration;
 
-use crate::configuration::{Configuration, LoadError, RefreshError};
-use crate::qpu::{ExecutionError, JobId};
-use crate::{qpu, qvm, ExecutionData};
+use qcs_api_client_common::configuration::LoadError;
+use qcs_api_client_common::ClientConfiguration;
+
+use crate::execution_data;
+use crate::qpu::client::Qcs;
+use crate::qpu::quilc::CompilerOpts;
+use crate::qpu::rewrite_arithmetic;
+use crate::qpu::runner::JobId;
+use crate::qpu::ExecutionError;
+use crate::{qpu, qvm};
+use quil_rs::program::ProgramError;
+use quil_rs::Program;
 
 /// The builder interface for executing Quil programs on QVMs and QPUs.
 ///
 /// # Example
 ///
 /// ```rust
+/// use qcs_api_client_common::ClientConfiguration;
 /// use qcs::{Executable, RegisterData};
 ///
 ///
@@ -29,7 +39,7 @@ use crate::{qpu, qvm, ExecutionData};
 ///
 /// #[tokio::main]
 /// async fn main() {
-///     let mut result = Executable::from_quil(PROGRAM).with_shots(4).execute_on_qvm().await.unwrap();
+///     let mut result = Executable::from_quil(PROGRAM).with_config(ClientConfiguration::default()).with_shots(4).execute_on_qvm().await.unwrap();
 ///     // We know it's i8 because we declared the memory as `BIT` in Quil.
 ///     // "ro" is the only source read from by default if you don't specify a .read_from()
 ///     let data = result.registers.remove("ro").expect("Did not receive ro data").into_i8().unwrap();
@@ -59,7 +69,9 @@ pub struct Executable<'executable, 'execution> {
     readout_memory_region_names: Option<Vec<&'executable str>>,
     params: Parameters,
     compile_with_quilc: bool,
-    config: Option<Arc<Configuration>>,
+    compiler_options: CompilerOpts,
+    config: Option<ClientConfiguration>,
+    client: Option<Arc<Qcs>>,
     qpu: Option<qpu::Execution<'execution>>,
     qvm: Option<qvm::Execution>,
 }
@@ -89,7 +101,9 @@ impl<'executable> Executable<'executable, '_> {
             readout_memory_region_names: None,
             params: Parameters::new(),
             compile_with_quilc: true,
+            compiler_options: CompilerOpts::default(),
             config: None,
+            client: None,
             qpu: None,
             qvm: None,
         }
@@ -109,6 +123,7 @@ impl<'executable> Executable<'executable, '_> {
     /// # Example
     ///
     /// ```rust
+    /// use qcs_api_client_common::ClientConfiguration;
     /// use qcs::Executable;
     ///
     /// const PROGRAM: &str = r#"
@@ -122,6 +137,7 @@ impl<'executable> Executable<'executable, '_> {
     /// #[tokio::main]
     /// async fn main() {
     ///     let mut result = Executable::from_quil(PROGRAM)
+    ///         .with_config(ClientConfiguration::default()) // Unnecessary if you have ~/.qcs/settings.toml
     ///         .read_from("first")
     ///         .read_from("second")
     ///         .execute_on_qvm()
@@ -165,6 +181,7 @@ impl<'executable> Executable<'executable, '_> {
     /// # Example
     ///
     /// ```rust
+    /// use qcs_api_client_common::ClientConfiguration;
     /// use qcs::Executable;
     ///
     /// const PROGRAM: &str = "DECLARE theta REAL[2]";
@@ -172,6 +189,7 @@ impl<'executable> Executable<'executable, '_> {
     /// #[tokio::main]
     /// async fn main() {
     ///     let mut exe = Executable::from_quil(PROGRAM)
+    ///         .with_config(ClientConfiguration::default()) // Unnecessary if you have ~/.qcs/settings.toml
     ///         .read_from("theta");
     ///     
     ///     for theta in 0..2 {
@@ -200,7 +218,7 @@ impl<'executable> Executable<'executable, '_> {
             .remove(&param_name)
             .unwrap_or_else(|| vec![0.0; index]);
 
-        if index + 1 > values.len() {
+        if index >= values.len() {
             values.resize(index + 1, 0.0);
         }
 
@@ -209,9 +227,17 @@ impl<'executable> Executable<'executable, '_> {
 
         self
     }
+
+    /// Set the default configuration to be used when constructing clients
+    #[must_use]
+    pub fn with_config(mut self, config: ClientConfiguration) -> Self {
+        self.config = Some(config);
+        self
+    }
 }
 
-type ExecuteResult = Result<ExecutionData, Error>;
+type ExecuteResultQVM = Result<execution_data::Qvm, Error>;
+type ExecuteResultQPU = Result<execution_data::Qpu, Error>;
 
 impl Executable<'_, '_> {
     /// Specify a number of times to run the program for each execution. Defaults to 1 run or "shot".
@@ -229,6 +255,13 @@ impl Executable<'_, '_> {
         self
     }
 
+    /// If set, the value will override the default compiler options
+    #[must_use]
+    pub fn compiler_options(mut self, options: CompilerOpts) -> Self {
+        self.compiler_options = options;
+        self
+    }
+
     fn get_readouts(&self) -> &[&str] {
         return self
             .readout_memory_region_names
@@ -250,8 +283,9 @@ impl Executable<'_, '_> {
     /// # Errors
     ///
     /// See [`Error`].
-    pub async fn execute_on_qvm(&mut self) -> ExecuteResult {
-        let config = self.get_config().await.unwrap_or_default();
+    pub async fn execute_on_qvm(&mut self) -> ExecuteResultQVM {
+        let config = self.get_config().await?;
+
         let mut qvm = if let Some(qvm) = self.qvm.take() {
             qvm
         } else {
@@ -261,29 +295,34 @@ impl Executable<'_, '_> {
             .run(self.shots, self.get_readouts(), &self.params, &config)
             .await;
         self.qvm = Some(qvm);
-        result.map_err(Error::from).map(|registers| ExecutionData {
-            registers,
-            duration: None,
-        })
+        result
+            .map_err(Error::from)
+            .map(|registers| execution_data::Qvm {
+                registers,
+                duration: None,
+            })
     }
 
-    /// Load `self.config` if not yet loaded, then return a reference to it.
-    async fn get_config(&mut self) -> Result<Arc<Configuration>, Error> {
+    async fn get_config(&mut self) -> Result<ClientConfiguration, Error> {
         if let Some(config) = &self.config {
             Ok(config.clone())
         } else {
-            let config = Arc::new(Configuration::load().await?);
+            let config = ClientConfiguration::load().await?;
             self.config = Some(config.clone());
             Ok(config)
         }
     }
 
-    /// Refresh `self.config` and return it.
-    async fn refresh_config(&mut self) -> Result<Arc<Configuration>, Error> {
-        let config = self.get_config().await?.as_ref().clone();
-        let refreshed = Arc::new(config.refresh().await?);
-        self.config = Some(refreshed.clone());
-        Ok(refreshed)
+    /// Load `self.client` if not yet loaded, then return a reference to it.
+    async fn get_client(&mut self) -> Result<Arc<Qcs>, Error> {
+        if let Some(client) = &self.client {
+            Ok(client.clone())
+        } else {
+            let config = self.get_config().await?;
+            let client = Arc::new(Qcs::with_config(config));
+            self.client = Some(client.clone());
+            Ok(client)
+        }
     }
 }
 
@@ -298,31 +337,16 @@ impl<'execution> Executable<'_, 'execution> {
                 return Ok(qpu);
             }
         }
-        let mut config = self.get_config().await?;
-        match qpu::Execution::new(
+        qpu::Execution::new(
             self.quil.clone(),
             self.shots,
             id,
-            config.clone(),
+            self.get_client().await?,
             self.compile_with_quilc,
+            self.compiler_options,
         )
         .await
-        {
-            Ok(qpu) => Ok(qpu),
-            Err(ExecutionError::Unauthorized) => {
-                config = self.refresh_config().await?;
-                qpu::Execution::new(
-                    self.quil.clone(),
-                    self.shots,
-                    id,
-                    config,
-                    self.compile_with_quilc,
-                )
-                .await
-                .map_err(Error::from)
-            }
-            Err(err) => Err(Error::from(err)),
-        }
+        .map_err(Error::from)
     }
 
     /// Compile the program and execute it on a QPU, waiting for results.
@@ -351,7 +375,10 @@ impl<'execution> Executable<'_, 'execution> {
     /// 1. Missing parameters that should be filled with [`Executable::with_parameter`]
     ///
     /// [quilc]: https://github.com/quil-lang/quilc
-    pub async fn execute_on_qpu(&mut self, quantum_processor_id: &'execution str) -> ExecuteResult {
+    pub async fn execute_on_qpu(
+        &mut self,
+        quantum_processor_id: &'execution str,
+    ) -> ExecuteResultQPU {
         let job_handle = self.submit_to_qpu(quantum_processor_id).await?;
         self.retrieve_results(job_handle).await
     }
@@ -368,28 +395,22 @@ impl<'execution> Executable<'_, 'execution> {
         &mut self,
         quantum_processor_id: &'execution str,
     ) -> Result<JobHandle<'execution>, Error> {
-        let mut qpu = self.qpu_for_id(quantum_processor_id).await?;
-        let mut config = self.get_config().await?;
+        let JobHandle {
+            job_id,
+            readout_map,
+            ..
+        } = self
+            .qpu_for_id(quantum_processor_id)
+            .await?
+            .submit(&self.params)
+            .await?;
 
-        let response = match qpu.submit(&self.params, self.get_readouts(), &config).await {
-            Ok(response) => Ok(response),
-            Err(ExecutionError::Unauthorized) => {
-                config = self.refresh_config().await?;
-                qpu.submit(&self.params, self.get_readouts(), &config).await
-            }
-            Err(err) => Err(err),
-        }
-        .map(|job_id| JobHandle {
+        // The unpacking/re-packing here is to avoid lifetime issues.
+        Ok(JobHandle {
             job_id,
             quantum_processor_id,
-        });
-        self.qpu = if let Err(ExecutionError::QpuUnavailable(_)) = response {
-            // Could mean maintenance which requires recompile
-            None
-        } else {
-            Some(qpu)
-        };
-        response.map_err(Error::from)
+            readout_map,
+        })
     }
 
     /// Wait for the results of a job submitted via [`Executable::submit_to_qpu`] to complete.
@@ -397,9 +418,12 @@ impl<'execution> Executable<'_, 'execution> {
     /// # Errors
     ///
     /// See [`Executable::execute_on_qpu`].
-    pub async fn retrieve_results(&mut self, job_handle: JobHandle<'execution>) -> ExecuteResult {
+    pub async fn retrieve_results(
+        &mut self,
+        job_handle: JobHandle<'execution>,
+    ) -> ExecuteResultQPU {
         let qpu = self.qpu_for_id(job_handle.quantum_processor_id).await?;
-        qpu.retrieve_results(job_handle.job_id)
+        qpu.retrieve_results(job_handle.job_id, job_handle.readout_map)
             .await
             .map_err(Error::from)
     }
@@ -439,8 +463,23 @@ pub enum Error {
     /// There was some problem with the provided Quil program. This could be a syntax error with
     /// quil, providing Quil-T to `quilc` or `qvm` (which is not supported), or forgetting to set
     /// some parameters.
+    #[error("There was a problem with the Quil program: {0}")]
+    Quil(#[from] ProgramError<Program>),
+    /// There was a problem when compiling the Quil program.
     #[error("There was a problem compiling the Quil program: {0}")]
     Compilation(String),
+    /// There was a problem when translating the Quil program.
+    #[error("There was a problem translating the Quil program: {0}")]
+    Translation(String),
+    /// There was a problem when rewriting parameter arithmetic in the Quil program.
+    #[error("There was a problem rewriting parameter arithmetic in the Quil program: {0}")]
+    RewriteArithmetic(#[from] rewrite_arithmetic::Error),
+    /// There was a problem when substituting parameters in the Quil program.
+    #[error("There was a problem substituting parameters in the Quil program: {0}")]
+    Substitution(String),
+    /// The Quil program is missing readout sources.
+    #[error("The Quil program is missing readout sources")]
+    MissingRoSources,
     /// This error returns when a runtime check that _should_ always pass fails. This most likely
     /// indicates a bug in the SDK and should be reported to
     /// [GitHub](https://github.com/rigetti/qcs-sdk-rust/issues),
@@ -451,6 +490,9 @@ pub enum Error {
     /// [`Executable::retrieve_results`] can invalidate the handle.
     #[error("The job handle was not valid")]
     InvalidJobHandle,
+    /// Occurs when failing to construct a [`QcsClient`].
+    #[error("The QCS client configuration failed to load")]
+    QcsConfigLoadFailure(#[from] LoadError),
 }
 
 #[derive(Debug, Copy, Clone, Eq, PartialEq)]
@@ -480,36 +522,18 @@ pub enum Service {
     Qpu,
 }
 
-impl From<LoadError> for Error {
-    fn from(err: LoadError) -> Self {
-        Self::Settings(format!("{}", err))
-    }
-}
-
-impl From<RefreshError> for Error {
-    fn from(err: RefreshError) -> Self {
-        match err {
-            RefreshError::NoRefreshToken => Self::Settings(String::from(
-                "No `refresh_token` was found in your QCS secrets file for the selected profile. \
-                    You can change profiles with the `QCS_PROFILE_NAME` environment variable.",
-            )),
-            RefreshError::FetchError(_) => Self::Authentication,
-        }
-    }
-}
-
 impl From<ExecutionError> for Error {
     fn from(err: ExecutionError) -> Self {
         match err {
-            ExecutionError::QpuNotFound => Self::QpuNotFound,
-            ExecutionError::QpuUnavailable(duration) => Self::QpuUnavailable(duration),
-            ExecutionError::Unauthorized => Self::Authentication,
-            ExecutionError::QcsCommunication => Self::Connection(Service::Qcs),
-            ExecutionError::Quil(message) => Self::Compilation(message),
             ExecutionError::Unexpected(inner) => Self::Unexpected(format!("{:?}", inner)),
             ExecutionError::Quilc { .. } => Self::Connection(Service::Quilc),
-            ExecutionError::Qcs(message) => Self::Unexpected(message),
-            ExecutionError::ProgramNotSubmitted => Self::InvalidJobHandle,
+            ExecutionError::QcsClient(v) => Self::Unexpected(format!("{:?}", v)),
+            ExecutionError::IsaError(v) => Self::Unexpected(format!("{:?}", v)),
+            ExecutionError::ReadoutParse(v) => Self::Unexpected(format!("{:?}", v)),
+            ExecutionError::Quil(e) => Self::Quil(e),
+            ExecutionError::Compilation { details } => Self::Compilation(details),
+            ExecutionError::RewriteArithmetic(e) => Self::RewriteArithmetic(e),
+            ExecutionError::Substitution(message) => Self::Substitution(message),
         }
     }
 }
@@ -529,24 +553,66 @@ impl From<qvm::Error> for Error {
 
 /// The result of calling [`Executable::submit_to_qpu`]. Represents a quantum program running on
 /// a QPU. Can be passed to [`Executable::retrieve_results`] to retrieve the results of the job.
-#[derive(Debug, Clone, Eq, PartialEq)]
+#[derive(Debug, Clone, PartialEq, Eq)]
 pub struct JobHandle<'executable> {
     job_id: JobId,
     quantum_processor_id: &'executable str,
+    readout_map: HashMap<String, String>,
 }
 
-impl JobHandle<'_> {
+impl<'a> JobHandle<'a> {
+    #[must_use]
+    pub(crate) fn new(
+        job_id: JobId,
+        quantum_processor_id: &'a str,
+        readout_map: HashMap<String, String>,
+    ) -> Self {
+        Self {
+            job_id,
+            quantum_processor_id,
+            readout_map,
+        }
+    }
+
     /// The string representation of the QCS Job ID. Useful for debugging.
     #[must_use]
     pub fn job_id(&self) -> &str {
         self.job_id.0.as_str()
     }
+
+    /// The readout map from source readout memory locations to the
+    /// filter pipeline node which publishes the data.
+    #[must_use]
+    pub fn readout_map(&self) -> &HashMap<String, String> {
+        &self.readout_map
+    }
+}
+
+#[cfg(test)]
+mod describe_get_config {
+    use qcs_api_client_openapi::common::ClientConfiguration;
+
+    use crate::Executable;
+
+    #[tokio::test]
+    async fn it_resizes_params_dynamically() {
+        let mut exe = Executable::from_quil("").with_config(ClientConfiguration::default());
+        let foo_len = |exe: &mut Executable| exe.params.get("foo").unwrap().len();
+
+        exe.with_parameter("foo", 0, 0.0);
+        assert_eq!(foo_len(&mut exe), 1);
+
+        exe.with_parameter("foo", 10, 10.0);
+        assert_eq!(foo_len(&mut exe), 11);
+    }
 }
 
 #[cfg(test)]
 #[cfg(feature = "manual-tests")]
 mod describe_qpu_for_id {
-    use super::*;
+    use std::sync::Arc;
+
+    use crate::{configuration::Configuration, qpu, Executable};
 
     #[tokio::test]
     async fn it_refreshes_auth_token() {
@@ -560,7 +626,7 @@ mod describe_qpu_for_id {
             panic!("Expected an error!");
         };
         let result_string = format!("{:?}", err);
-        assert!(result_string.contains("refresh token"))
+        assert!(result_string.contains("refresh_token"));
     }
 
     #[tokio::test]
@@ -572,7 +638,7 @@ mod describe_qpu_for_id {
             qpu::Execution::new(
                 "".into(),
                 shots,
-                "Aspen-9",
+                "Aspen-11",
                 exe.get_config().await.unwrap_or_default(),
                 exe.compile_with_quilc,
             )
@@ -582,7 +648,7 @@ mod describe_qpu_for_id {
         // Load config with no credentials to prevent creating a new Execution if it tries
         exe.config = Some(Arc::new(Configuration::default()));
 
-        assert!(exe.qpu_for_id("Aspen-9").await.is_ok());
+        assert!(exe.qpu_for_id("Aspen-11").await.is_ok());
     }
 
     #[tokio::test]
@@ -622,17 +688,19 @@ mod describe_qpu_for_id {
 
 #[cfg(test)]
 #[cfg(feature = "manual-tests")]
-mod describe_take_or_load_config {
-    use super::*;
+mod describe_get_config {
+    use std::sync::Arc;
+
+    use crate::{configuration::Configuration, Executable};
 
     #[tokio::test]
     async fn it_returns_cached_values() {
         let mut exe = Executable::from_quil("");
-        let mut config = Arc::new(Configuration::default());
+        let mut config = Configuration::default();
         config.quilc_url = String::from("test");
+        let config = Arc::new(config);
         exe.config = Some(config.clone());
-        let gotten = exe.take_or_load_config().await;
+        let gotten = exe.get_config().await.unwrap_or_default();
         assert_eq!(gotten.quilc_url, config.quilc_url);
-        assert!(matches!(exe.config, None));
     }
 }
